#!/usr/bin/env python
#
#  ninix-install.py - a command-line installer for ninix
#                     Tamito KAJIYAMA <22 June 2001>
#
#  $Id: ninix-install.py,v 1.23 2003/07/23 00:08:49 shy Exp $
#

# env.var: NINIX_HOME, NINIX_USER, NINIX_ARCHIVE, TMPDIR
# command: cp, mv, rm, mkdir, lha, zip, find, chmod


import getopt
import os
import re
import socket
import string
import sys
import tempfile
import urllib

if os.environ.has_key("DISPLAY"):
    del os.environ["DISPLAY"]

import ninix.main
import ninix.alias
import ninix.config
import ninix.home

PROGRAM_NAME = os.path.basename(sys.argv[0])

USAGE = """\
Usage: %s [options] [file ...]
Options:
  -g, --ghost            assume that all files are ghosts
  -s, --shell            assume that all files are shells
  -b, --balloon          assume that all files are balloons
  -P, --plugin           assume that all files are plug-ins
  -S, --supplement NAME  the name of the supplement target
  -i, --interactive      prompt before overwrite
  -d, --download         download files into the archive directory
  -H, --homedir DIR      ninix home directory (default: ~/.ninix)
  -U, --userdir DIR      user data directory (default: ninix home directory)
  -A, --arcdir DIR       archive directory (default: ~/.ninix/archive)
  -T, --tempdir DIR      temporary directory (default: /usr/tmp)
  -r, --reload           send "reload" request to local SSTP server
  -p, --port NUM         port number for SSTP connection (default: 9801)
  -L, --lower            lower file names with upper case letters
  -h, --help             show this message
"""

def usage():
    sys.stderr.write(USAGE % PROGRAM_NAME)
    sys.exit(1)

class InstallError(Exception):
    pass

def fatal(error):
    raise InstallError, error

# mode masks
INTERACTIVE = 1
DOWNLOAD    = 2

def main():
    global mode, arcdir
    try:
        options, files = getopt.getopt(
            sys.argv[1:], "gsbS:PRridxH:U:A:T:rp:q:Lh",
            ["ghost", "shell", "balloon", "supplement=", "plugin",
             "homedir=", "userdir=", "arcdir=", "tempdir=",
             "interactive", "download", "reload", "port=",
             "lower", "help"])
    except getopt.error, e:
        sys.stderr.write("Error: %s\n" % str(e))
        usage()
    homedir = os.environ.get("NINIX_HOME")
    userdir = os.environ.get("NINIX_USER")
    arcdir = os.environ.get("NINIX_ARCHIVE", "~/.ninix/archive")
    tempdir = None
    filetype = None
    target = None
    mode = 0
    to_lower = 0
    to_reload = 0
    host = ""
    port = 9801
    for opt, val in options:
        if opt in ["-h", "--help"]:
            usage()
        elif opt in ["-g", "--ghost"]:
            filetype = "ghost"
        elif opt in ["-s", "--shell"]:
            filetype = "shell"
        elif opt in ["-b", "--balloon"]:
            filetype = "balloon"
        elif opt in ["-P", "--plugin"]:
            filetype = "plugin"
        elif opt in ["-S", "--supplement"]:
            filetype = "supplement"
            target = val
        elif opt in ["-H", "--homedir"]:
            homedir = val
        elif opt in ["-U", "--userdir"]:
            userdir = val
        elif opt in ["-A", "--arcdir"]:
            arcdir = val
        elif opt in ["-T", "--tempdir"]:
            tempdir = val
        elif opt in ["-i", "--interactive"]:
            mode = mode | INTERACTIVE
        elif opt in ["-d", "--download"]:
            mode = mode | DOWNLOAD
        elif opt in ["-r", "--reload"]:
            to_reload = 1
        elif opt in ["-p", "--port"]:
            port = int(val)
        elif opt in ["-L", "--lower"]:
            to_lower = 1
    if homedir:
        ninix.home.NINIX_HOME = homedir
    if userdir:
        ninix.home.NINIX_USER = userdir
    if tempdir:
        tempfile.tempdir = os.path.expanduser(tempdir)
    arcdir = os.path.expanduser(arcdir)
    if not os.path.exists(arcdir):
        run("mkdir -p %s" % escape(arcdir))
    if to_lower:
        n = 0
        homedir = ninix.home.get_ninix_home()
        for data in ["ghost", "balloon"]:
            try:
                buffer = os.listdir(os.path.join(homedir, data))
            except OSError:
                continue
            for subdir in buffer:
                path = os.path.join(homedir, data, subdir)
                if os.path.isdir(path):
                    n = n + lower_files(path)
        if n > 0:
            print n, "files renamed"
    buffer = []
    for filename in files:
        if filename[:1] == "@":
            try:
                lines = open(filename[1:]).readlines()
            except IOError, e:
                sys.stderr.write("cannot read '%s' (skipped)\n" % filename[1:])
            else:
                for line in map(string.strip, lines):
                    if line and line[0] != "#":
                        buffer.append(line)
        else:
            buffer.append(filename)
    for filename in buffer:
        try:
            install(filename, filetype, target, ninix.home.get_ninix_home())
        except InstallError, error:
            sys.stderr.write("%s: %s (skipped)\n" % (filename, error))
    if to_reload:
        reload(host, port)

def escape(s):
    return '"' + string.replace(s, '"', '\\"') + '"'

def install(filename, filetype, target, homedir):
    # check archive format
    suffix = string.lower(filename[-4:])
    if suffix in [".nar", ".zip"]:
        archiver = "unzip -o %(f)s -d %(d)s"
    elif suffix == ".lzh":
        archiver = "lha xfw=%(d)s %(f)s"
    else:
        fatal("unknown archive format")
    # extract files from the archive
    tmpdir = tempfile.mktemp("ninix")
    if not run("mkdir -p %s" % escape(tmpdir)):
        fatal("cannot make temporary directory")
    url = None
    if filename[:5] == "http:" or filename[:4] == "ftp:":
        url = filename
        filename = download(filename, tmpdir)
        if filename is None:
            os.system("rm -rf %s" % escape(tmpdir))
            fatal("cannot download the archive file")
    if not run(archiver % {"f": escape(filename), "d": escape(tmpdir)}):
        os.system("rm -rf %s" % escape(tmpdir))
        fatal("cannot extract files from the archive")
    if url and not mode & DOWNLOAD:
        run("rm %s" % escape(filename))
    os.system("find %s -type d -exec chmod u+rwx {} \\;" % escape(tmpdir))
    os.system("find %s -type f -exec chmod u+rw  {} \\;" % escape(tmpdir))
    rename_files(tmpdir)
    # check the file type (ghost, shell, balloon or supplement)
    error = None
    inst = ninix.home.read_install_txt(tmpdir)
    if inst:
        guess = inst.get("type")
        if guess == "ghost":
            if os.path.exists(os.path.join(tmpdir, "ghost", "master")):
                guess = "ghost.redo"
            else:
                guess = "ghost.inverse"
        elif guess == "ghost with balloon":
            guess = "ghost.inverse"
        elif guess in ["shell", "shell with balloon"]:
            if inst.has_key("accept"):
                guess = "supplement"
            else:
                guess = "shell.inverse"
    else:
        guess = None
        for f, t in [("ghost/master/", "ghost.redo"),
                     ("shell/",        "shell.redo"),
                     ("kawari.ini",    "ghost.inverse"),
                     ("ai.txt",        "ghost.inverse"),
                     ("ai.dtx",        "ghost.inverse"),
                     ("surface/",      "shell.inverse"),
                     ("balloons0.png", "balloon"),
                     ("balloons0.jpg", "balloon"),
                     ("plugin.txt",    "plugin")]:
            if os.path.exists(os.path.join(tmpdir, f)):
                guess = t
                break
    if filetype is None and guess is None:
        error = "cannot guess the file type; use option -g, -s, -b, -P or -S"
    elif filetype is None:
        filetype = guess
    elif filetype == "ghost" and guess in ["ghost.redo", "ghost.inverse"]:
        filetype = guess
    elif filetype == "shell" and guess in ["ghost.redo", "shell.redo", "supplement"]:
        filetype = "shell.redo"
    elif filetype == "shell" and guess == "ghost.inverse":
        filetype = "shell.inverse"
    elif guess is not None and filetype != guess:
        error = "wrong file type option; correct file type is " + guess
    try:
        if error is not None:
            fatal(error)
        elif filetype == "ghost.redo":
            install_redo_ghost(filename, tmpdir, homedir)
        elif filetype == "shell.redo":
            install_redo_shell(filename, tmpdir, homedir)
        elif filetype == "supplement":
            install_redo_supplement(filename, tmpdir, homedir, target)
        elif filetype == "ghost.inverse":
            install_inverse_ghost(filename, tmpdir, homedir)
        elif filetype == "shell.inverse":
            install_inverse_shell(filename, tmpdir, homedir)
        elif filetype == "balloon":
            install_balloon(filename, tmpdir, homedir)
        elif filetype == "plugin":
            install_plugin(filename, tmpdir, homedir)
    finally:
        run("rm -rf %s" % escape(tmpdir))

class URLopener(urllib.FancyURLopener):
    version = "ninix/%s" % ninix.main.VERSION
urllib._urlopener = URLopener()

def download(url, dir):
    print "downloading", url
    try:
        ifile = urllib.urlopen(url)
    except IOError:
        return None
    headers = ifile.info()
    if headers.has_key("content-length"):
        print "(size = %s bytes)" % headers.get("content-length")
    if mode & DOWNLOAD:
        dir = arcdir
    filename = os.path.join(dir, os.path.basename(url))
    try:
        ofile = open(filename, "w")
    except IOError:
        return None
    while 1:
        data = ifile.read(4096)
        if not data:
            break
        ofile.write(data)
    ifile.close()
    ofile.close()
    # check the format of the downloaded file
    suffix = string.lower(filename[-4:])
    if suffix in [".nar", ".zip"]:
        archiver = "unzip -t"
    elif suffix == ".lzh":
        archiver = "lha t"
    else:
        fatal("unknown archive format")
    if os.system("%s %s >/dev/null 2>&1" % (archiver, escape(filename))):
        return None
    return filename

def rename_files(dir):
    for filename in os.listdir(dir):
        filename2 = string.lower(filename)
        path = os.path.join(dir, filename2)
        if filename != filename2:
            os.rename(os.path.join(dir, filename), path)
        if os.path.isdir(path):
            rename_files(path)

def list_all_directories(top, dir):
    dirlist = []
    for path in os.listdir(os.path.join(top, dir)):
        if os.path.isdir(os.path.join(top, dir, path)):
            dirlist.extend(list_all_directories(top, os.path.join(dir, path)))
            dirlist.append(os.path.join(dir, path))
    return dirlist

def remove_files_and_dirs(mask, dir):
    path = os.path.abspath(dir)
    if not os.path.isdir(path):
        return
    os.path.walk(path, remove_files, mask)
    list = list_all_directories(path, '')
    list.sort()
    list.reverse()
    for name in list:
        current_path = os.path.join(path, name)
        if os.path.isdir(current_path):
            head, tail = os.path.split(current_path)
            if not tail in mask and not os.listdir(current_path):
                run("rm -rf %s" % escape(current_path))

def remove_files(mask, dir, list):
    for name in list:
        path = os.path.join(dir, name)
        if os.path.isdir(path) or name in mask:
            pass
        else:
            run("rm -f %s" % escape(path))

def lower_files(dir):
    n = 0
    for filename in os.listdir(dir):
        filename2 = string.lower(filename)
        path = os.path.join(dir, filename2)
        if filename != filename2:
            os.rename(os.path.join(dir, filename), path)
            print "renamed", os.path.join(dir, filename)
            n = n + 1
        if os.path.isdir(path):
            n = n + lower_files(path)
    return n

def confirm_overwrite(path):
    if not mode & INTERACTIVE:
        return 1
    print PROGRAM_NAME + ': overwrite "%s"? (yes/no)' % path
    try:
        answer = raw_input()
    except EOFError:
        answer = None
    except KeyboardInterrupt:
        sys.exit(1)
    if not answer:
        return 0
    return string.lower(answer[0]) == "y"

def confirm_removal(path):
    if not mode & INTERACTIVE:
        return 1
    print PROGRAM_NAME + ': remove "%s"? (yes/no)' % path
    try:
        answer = raw_input()
    except EOFError:
        answer = None
    except KeyboardInterrupt:
        sys.exit(1)
    if not answer:
        return 0
    return string.lower(answer[0]) == "y"

def uninstall_plugin(homedir, name):
    try:
        dirlist = os.listdir(os.path.join(homedir, "plugin"))
    except OSError:
        return
    for subdir in dirlist:
        path = os.path.join(homedir, "plugin", subdir)
        plugin = ninix.home.read_plugin_txt(path)
        if plugin is None:
            continue
        plugin_name, plugin_dir, startup, menu_items = plugin
        if plugin_name == name:
            plugin_dir = os.path.join(homedir, "plugin", subdir)
            if confirm_removal(plugin_dir):
                run("rm -rf %s" % escape(plugin_dir))

def install_redo_ghost(archive, tmpdir, homedir):
    global mode
    # find install.txt
    inst = ninix.home.read_install_txt(tmpdir)
    if inst is None:
        fatal("install.txt not found")
    dir = inst.get("directory")
    if dir is None:
        fatal("\"directory\" not found in install.txt")
    dir = dir.encode('utf-8')
    prefix = os.path.join(homedir, "ghost", dir)
    ghost_src = os.path.join(tmpdir, "ghost", "master")
    shell_src = os.path.join(tmpdir, "shell")
    ghost_dst = os.path.join(prefix, "ghost", "master")
    shell_dst = os.path.join(prefix, "shell")
    filelist = []
    filelist.append((os.path.join(tmpdir, "install.txt"),
                     os.path.join(prefix, "install.txt")))
    # find ghost/master/descript.txt
    desc = ninix.home.read_descript_txt(ghost_src)
    if desc:
        filelist.append((os.path.join(ghost_src, "descript.txt"),
                         os.path.join(ghost_dst, "descript.txt")))
    # find ghost/master/alias.txt
    path = os.path.join(ghost_src, "alias.txt")
    if os.path.exists(path):
        filelist.append((path, os.path.join(ghost_dst, "alias.txt")))
        aliases = ninix.alias.open(path)
    else:
        aliases = {}
    # find pseudo AI
    pseudo_ai = find_pseudo_ai(ghost_src, ghost_dst)
    if pseudo_ai is None:
        fatal("pseudo AI not found - cannot be used as a ghost")
    filelist.extend(pseudo_ai)
    # XXX: ad hoc work around for Mayura v3.10
    path = os.path.join(ghost_src, "makotob.dll")
    if os.path.exists(path):
        os.system("mv %s %s" % (escape(path),
                                escape(os.path.join(ghost_src, "makoto.dll"))))
    # find makoto.dll
    for filename in aliases.get("makoto", ["makoto.dll"]):
        path = os.path.join(ghost_src, filename)
        try:
            data = open(path).read()
        except IOError:
            continue
        if string.find(data, "Makoto Basic with Select and Repeat") > 0:
            os.system("rm -f %s" % escape(path))
            os.system("touch %s" % escape(path))
            filelist.append((path, os.path.join(ghost_dst, filename)))
    # find shell
    shell = find_redo_shell(shell_src, shell_dst)
    if shell:
        filelist.extend(shell)
    # find balloon
    balloon_dir = inst and inst.get("balloon.directory")
    if balloon_dir:
        balloon_dir = string.lower(balloon_dir.encode('utf-8'))
        filelist.extend(find_balloon(os.path.join(tmpdir,    balloon_dir),
                                     os.path.join(ghost_dst, balloon_dir)))
    if os.path.exists(prefix):
        inst_dst = ninix.home.read_install_txt(prefix)
        if not inst_dst or inst_dst.get("name") != inst.get("name"):
            mode  = mode | INTERACTIVE
        if inst.getint("refresh", 0):
            # uninstall older versions of the ghost
            if confirm_removal(prefix):
                mask = string.split(string.lower(inst.get("refreshundeletemask", "")), ':')
                mask.append('HISTORY')
                remove_files_and_dirs(mask, prefix)
            else:
                return
        else:
            if not confirm_overwrite(prefix):
                return
    # install files
    print "installing", archive, "(ghost)"
    install_files(filelist)

def install_redo_shell(archive, tmpdir, homedir):
    global mode
    # find install.txt
    inst = ninix.home.read_install_txt(tmpdir)
    if inst is None:
        fatal("install.txt not found")
    dir = inst.get("directory")
    if dir is None:
        fatal("\"directory\" not found in install.txt")
    dir = dir.encode('utf-8')
    prefix = os.path.join(homedir, "ghost", dir)
    ghost_src = os.path.join(tmpdir, "ghost", "master")
    shell_src = os.path.join(tmpdir, "shell")
    ghost_dst = os.path.join(prefix, "ghost", "master")
    shell_dst = os.path.join(prefix, "shell")
    filelist = []
    filelist.append((os.path.join(tmpdir, "install.txt"),
                     os.path.join(prefix, "install.txt")))
    # find ghost/master/descript.txt
    desc = ninix.home.read_descript_txt(ghost_src)
    if desc:
        filelist.append((os.path.join(ghost_src, "descript.txt"),
                         os.path.join(ghost_dst, "descript.txt")))
    # find shell
    shell = find_redo_shell(shell_src, shell_dst)
    if shell:
        filelist.extend(shell)
    # find balloon
    balloon_dir = inst and inst.get("balloon.directory")
    if balloon_dir:
        balloon_dir = string.lower(balloon_dir.encode('utf-8'))
        filelist.extend(find_balloon(os.path.join(tmpdir,    balloon_dir),
                                     os.path.join(ghost_dst, balloon_dir)))
    if os.path.exists(prefix):
        inst_dst = ninix.home.read_install_txt(prefix)
        if not inst_dst or inst_dst.get("name") != inst.get("name"):
            mode  = mode | INTERACTIVE
        if inst.getint("refresh", 0):
            # uninstall older versions of the ghost
            if confirm_removal(prefix):
                mask = string.split(string.lower(inst.get("refreshundeletemask", "")), ':')
                mask.append('HISTORY')
                remove_files_and_dirs(mask, prefix)
            else:
                return
        else:
            if not confirm_overwrite(prefix):
                return
    # install files
    print "installing", archive, "(shell)"
    install_files(filelist)

def install_redo_supplement(archive, tmpdir, homedir, target):
    inst = ninix.home.read_install_txt(tmpdir)
    if inst and inst.has_key("accept"):
        print "searching supplement target ...",
        candidates = []
        try:
            dirlist = os.listdir(os.path.join(homedir, "ghost"))
        except OSError:
            dirlist = []
        for dirname in dirlist:
            path = os.path.join(homedir, "ghost", dirname)
            if os.path.exists(os.path.join(path, "shell", "surface.txt")):
                continue
            desc = ninix.home.read_descript_txt(
                os.path.join(path, "ghost", "master"))
            if desc and desc.get("sakura.name") == inst.get("accept"):
                candidates.append(dirname)
        if not candidates:
            print "not found"
        elif target and target not in candidates:
            print '"%s" not found' % target
            return
        elif len(candidates) == 1 or target in candidates:
            if target:
                path = os.path.join(homedir, "ghost", target)
            else:
                path = os.path.join(homedir, "ghost", candidates[0])
            if inst.has_key("directory"):
                if inst.get("type") == "shell":
                    path = os.path.join(path, "shell", inst["directory"])
                else:
                    if not inst.has_key("type"):
                        print "supplement type not specified"
                    else:
                        print "unsupported supplement type:", inst["type"]
                    return
            print "found"
            if not os.path.exists(path):
                run("mkdir -p %s" % escape(path))
            run("rm -f %s" % escape(os.path.join(tmpdir, "install.txt")))
            run("cp -r %s/* %s" % (escape(tmpdir), escape(path)))
            return
        else:
            print "multiple candidates found"
            for candidate in candidates:
                print "   ", candidate
            fatal("try -S option with a target ghost name")
    # install supplement as a stand alone shell
    install_redo_shell(archive, tmpdir, homedir)

def find_redo_shell(srcdir, dstdir):
    if not os.path.exists(srcdir):
        return None
    filelist = []
    for dirname in os.listdir(srcdir):
        shell_src = os.path.join(srcdir, dirname)
        shell_dst = os.path.join(dstdir, dirname)
        if not os.path.isdir(shell_src):
            continue
        path = os.path.join(shell_src, "descript.txt")
        if os.path.exists(path):
            filelist.append((path, os.path.join(shell_dst, "descript.txt")))
        path = os.path.join(shell_src, "surfaces.txt")
        if os.path.exists(path):
            filelist.append((path, os.path.join(shell_dst, "surfaces.txt")))
            for key, config in ninix.home.read_surfaces_txt(shell_src):
                if key[:7] == "surface":
                    for e in find_surface_elements(shell_src, shell_dst, config):
                        if e not in filelist:
                            filelist.append(e)
        path = os.path.join(shell_src, "alias.txt")
        if os.path.exists(path):
            filelist.append((path, os.path.join(shell_dst, "alias.txt")))
        filelist.extend(find_surface_set(shell_src, shell_dst))
    return filelist

def find_surface_elements(shell_src, shell_dst, config):
    filelist = []
    for key, method, filename, x, y in ninix.home.list_surface_elements(config):
        filename = string.lower(filename)
        path = os.path.join(shell_src, filename)
        basename, suffix = os.path.splitext(filename)
        if not os.path.exists(path):
            path = os.path.join(shell_src, basename + '.dgp')
            if os.path.exists(path):
                filename = basename + '.dgp'
            else:
                print "%s file not found: %s" % (key, filename)
                continue
        if suffix != ".png":
            print "unsupported image format for %s: %s" % (key, filename)
            continue
        filelist.append((path, os.path.join(shell_dst, filename)))
    return filelist

def install_inverse_ghost(archive, tmpdir, homedir):
    global mode
    # find install.txt
    inst = ninix.home.read_install_txt(tmpdir)
    if inst is None:
        fatal("install.txt not found")
    dir = inst.get("directory")
    if dir is None:
        fatal("\"directory\" not found in install.txt")
    dir = dir.encode('utf-8')
    prefix = os.path.join(homedir, "ghost", dir)
    ghost_dst = os.path.join(prefix, "ghost", "master")
    shell_dst = os.path.join(prefix, "shell")
    filelist = []
    filelist.append((os.path.join(tmpdir, "install.txt"),
                     os.path.join(prefix, "install.txt")))
    # find descript.txt
    desc = ninix.home.read_descript_txt(tmpdir)
    if desc:
        filelist.append((os.path.join(tmpdir,    "descript.txt"),
                         os.path.join(ghost_dst, "descript.txt")))
    # find pseudo AI dictionaries
    pseudo_ai = find_pseudo_ai(tmpdir, ghost_dst)
    if pseudo_ai is None:
        fatal("pseudo AI not found - cannot be used as a ghost")
    filelist.extend(pseudo_ai)
    # find makoto.dll
    path = os.path.join(tmpdir, "makoto.dll")
    try:
        data = open(path).read()
    except IOError:
        pass
    else:
        if string.find(data, "Makoto Basic with Select and Repeat") > 0:
            os.system("rm -f %s" % escape(path))
            os.system("touch %s" % escape(path))
            filelist.append((path, os.path.join(ghost_dst, "makoto.dll")))
    # find shell
    shell = find_inverse_shell(tmpdir, shell_dst)
    if shell:
        filelist.extend(shell)
    # find balloon
    balloon_dir = inst and inst.get("balloon.directory")
    if balloon_dir:
        balloon_dir = string.lower(balloon_dir.encode('utf-8'))
        filelist.extend(find_balloon(os.path.join(tmpdir,    balloon_dir),
                                     os.path.join(ghost_dst, balloon_dir)))
    if os.path.exists(prefix):
        inst_dst = ninix.home.read_install_txt(prefix)
        if not inst_dst or inst_dst.get("name") != inst.get("name"):
            mode  = mode | INTERACTIVE
        if inst.getint("refresh", 0):
            # uninstall older versions of the ghost
            if confirm_removal(prefix):
                mask = string.split(string.lower(inst.get("refreshundeletemask", "")), ':')
                mask.append('HISTORY')
                remove_files_and_dirs(mask, prefix)
            else:
                return
        else:
            if not confirm_overwrite(prefix):
                return
    # install files
    print "installing", archive, "(ghost)"
    install_files(filelist)

def install_inverse_shell(archive, tmpdir, homedir):
    global mode
    # find install.txt
    inst = ninix.home.read_install_txt(tmpdir)
    if inst is None:
        fatal("install.txt not found")
    dir = inst.get("directory")
    if dir is None:
        fatal("\"directory\" not found in install.txt")
    dir = dir.encode('utf-8')
    prefix = os.path.join(homedir, "ghost", dir)
    ghost_dst = os.path.join(prefix, "ghost", "master")
    shell_dst = os.path.join(prefix, "shell")
    filelist = []
    filelist.append((os.path.join(tmpdir, "install.txt"),
                     os.path.join(prefix, "install.txt")))
    # find descript.txt
    desc = ninix.home.read_descript_txt(tmpdir)
    if desc:
        filelist.append((os.path.join(tmpdir,    "descript.txt"),
                         os.path.join(ghost_dst, "descript.txt")))
    # find shell
    shell = find_inverse_shell(tmpdir, shell_dst)
    if shell:
        filelist.extend(shell)
    # find balloon
    balloon_dir = inst and inst.get("balloon.directory")
    if balloon_dir:
        balloon_dir = string.lower(balloon_dir.encode('utf-8'))
        filelist.extend(find_balloon(os.path.join(tmpdir,    balloon_dir),
                                     os.path.join(ghost_dst, balloon_dir)))
    if os.path.exists(prefix):
        inst_dst = ninix.home.read_install_txt(prefix)
        if not inst_dst or inst_dst.get("name") != inst.get("name"):
            mode  = mode | INTERACTIVE
        if inst.getint("refresh", 0):
            # uninstall older versions of the shell
            if confirm_removal(prefix):
                mask = string.split(string.lower(inst.get("refreshundeletemask", "")), ':')
                mask.append('HISTORY')
                remove_files_and_dirs(mask, prefix)
            else:
                return
        else:
            if not confirm_overwrite(prefix):
                return
    # install files
    print "installing", archive, "(shell)"
    install_files(filelist)

def find_inverse_shell(srcdir, dstdir):
    filelist = []
    # find descript.txt
    path = os.path.join(srcdir, "descript.txt")    
    if os.path.exists(path):
        descript_txt = path
    else:
        descript_txt = None
    # find alias.txt
    path = os.path.join(srcdir, "alias.txt")
    if os.path.exists(path):
        alias_txt = path
    else:
        alias_txt = None
    # find surface sets
    dirlist = []
    path = os.path.join(srcdir, "surface.txt")
    if os.path.exists(path):
        filelist.append((path, os.path.join(dstdir, "surface.txt")))
        config = ninix.config.open(path)
        for name, subdir in config.itemlist:
            dirlist.append((name, string.lower(subdir)))
    else:
        dirlist.append((None, "surface"))
    for name, subdir in dirlist:
        shell_src = os.path.join(srcdir, subdir)
        shell_dst = os.path.join(dstdir, subdir)
        if not os.path.isdir(shell_src):
            continue
        if descript_txt:
            # make a copy
            filelist.append((descript_txt,
                             os.path.join(shell_dst, "descript.txt")))
        if alias_txt:
            # make a copy
            filelist.append((alias_txt,
                             os.path.join(shell_dst, "alias.txt")))
        filelist.extend(find_surface_set(shell_src, shell_dst))
    return filelist

def install_balloon(archive, srcdir, homedir):
    global mode
    filelist = []
    # find install.txt
    inst = ninix.home.read_install_txt(srcdir)
    if inst is None:
        fatal("install.txt not found")
    #dstdir = os.path.join(homedir, "balloon", os.path.basename(archive)[:-4])
    dir = inst.get("directory")
    if dir is None:
        fatal("\"directory\" not found in install.txt")
    dir = dir.encode('utf-8')
    dstdir = os.path.join(homedir, "balloon", dir)
    # find balloon
    balloon = find_balloon(srcdir, dstdir)
    if not balloon:
        fatal("balloon not found")
    filelist.extend(balloon)
    filelist.append((os.path.join(srcdir, "install.txt"),
                     os.path.join(dstdir, "install.txt")))
    if os.path.exists(dstdir):
        inst_dst = ninix.home.read_install_txt(dstdir)
        if not inst_dst or inst_dst.get("name") != inst.get("name"):
            mode  = mode | INTERACTIVE
        if inst.getint("refresh", 0):
            # uninstall older versions of the balloon
            if confirm_removal(dstdir):
                mask = string.split(string.lower(inst.get("refreshundeletemask", "")), ':')
                remove_files_and_dirs(mask, dstdir)
            else:
                return
        else:
            if not confirm_overwrite(dstdir):
                return
    # install files
    print "installing", archive, "(balloon)"
    install_files(filelist)

def install_plugin(archive, srcdir, homedir):
    filelist = []
    dstdir = os.path.join(homedir, "plugin", os.path.basename(archive)[:-4])
    # find plugin.txt
    plugin = ninix.home.read_plugin_txt(srcdir)
    if plugin is None:
        fatal("failed to read plugin.txt")
    plugin_name, plugin_dir, startup, menu_items = plugin
    # find files
    for filename in os.listdir(srcdir):
        path = os.path.join(srcdir, filename)
        if os.path.isfile(path):
            filelist.append((path, os.path.join(dstdir, filename)))
    # uninstall older versions of the plugin
    uninstall_plugin(homedir, plugin_name)
    # install files
    print "installing", archive, "(plugin)"
    install_files(filelist)

def list_all_files(top, dir):
    filelist = []
    for path in os.listdir(os.path.join(top, dir)):
        if os.path.isdir(os.path.join(top, dir, path)):
            filelist.extend(list_all_files(top, os.path.join(dir, path)))
        else:
            filelist.append(os.path.join(dir, path))
    return filelist

def find_pseudo_ai(srcdir, dstdir): # redo, inverse
    # FIXME: This is kluge.
    filelist = []
    for path in list_all_files(srcdir, ''):
        filelist.append((os.path.join(srcdir, path),
                         os.path.join(dstdir, path)))
    return filelist

re_surface_img = re.compile(r"surface[0-9]+\.(png|dgp|pna)")
re_surface_txt = re.compile(r"surface[0-9]+[sa]\.txt")

def find_surface_set(srcdir, dstdir):
    filename_alias = {}
    path = os.path.join(srcdir, "alias.txt")
    if os.path.exists(path):
        dict = ninix.alias.open(path)
        for basename, alias in dict.items():
            if basename[:7] == "surface":
                filename_alias[alias] = basename
    filelist = []
    for filename in os.listdir(srcdir):
        if re_surface_img.match(filename):
            filelist.append((os.path.join(srcdir, filename),
                             os.path.join(dstdir, filename)))
            continue
        elif re_surface_txt.match(filename):
            filelist.append((os.path.join(srcdir, filename),
                             os.path.join(dstdir, filename)))
            continue
        basename, suffix = os.path.splitext(filename)
        if not (filename_alias.has_key(basename) and suffix in [".png", ".dgp", ".pna"]):
            continue
        filelist.append((os.path.join(srcdir, filename),
                         os.path.join(dstdir, filename)))
        for suffix in ["s.txt", "a.txt"]:
            filename = basename + suffix
            if os.path.exists(os.path.join(srcdir, filename)):
                filelist.append((os.path.join(srcdir, filename),
                                 os.path.join(dstdir, filename)))
    return filelist

re_balloon_img = re.compile(r"(balloon[skc][0-9]+|arrow[01]|sstp)\.(png|pna)")
re_balloon_txt = re.compile(r"balloon[sk][0-9]+s\.txt")

def find_balloon(srcdir, dstdir): # redo, inverse
    filelist = []
    if not os.path.exists(srcdir):
        return filelist
    path = os.path.join(srcdir, "descript.txt")
    if os.path.exists(path):
        filelist.append((path, os.path.join(dstdir, "descript.txt")))
    for filename in os.listdir(srcdir):
        if re_balloon_img.match(filename):
            basename, suffix = os.path.splitext(filename)
            filelist.append((
                os.path.join(srcdir, filename),
                os.path.join(dstdir, basename + suffix)))
        elif re_balloon_txt.match(filename):
            filelist.append((
                os.path.join(srcdir, filename),
                os.path.join(dstdir, filename)))
    return filelist

def install_files(filelist):
    for from_path, to_path in filelist:
        dirname, filename = os.path.split(to_path)
        if not os.path.exists(dirname):
            run("mkdir -p %s" % escape(dirname))
        run("cp %s %s" % (escape(from_path), escape(to_path)))

def reload(host, port):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((host, port))
    except socket.error:
        return
    s.send("EXECUTE SSTP/1.5\r\n"
           "Command: Reload\r\n"
           "Sender: ninix-aya/%s\r\n"
           "\r\n" % ninix.main.VERSION)
    s.recv(1024)
    s.close()

def run(command):
    print command
    if os.system(command):
        print "***FAILED***"
        return 0
    return 1

def readable(path):
    try:
        open(path).read(64)
    except IOError:
        return 0
    return 1

if __name__ == "__main__":
    main()
